import torch
import torch.nn as nn
import torch.nn.functional as F
from torchdiffeq import odeint
from typing import Optional
from typing_extensions import Literal

from .module import LatentODE, Encoder, Decoder


class TNODE(nn.Module):
    """
    Class to automatically infer treatment wise cellular dynamics using autoencoder and neural ODE.

    Parameters
    ----------
    n_int
        Number of genes.
    n_latent
        Dimension of latent space.
        (Default: 20)
    n_sample
        Number of samples or treatment as independant ODEs
        (Default: 1)
    n_ode_hidden
        The dimensionality of the hidden layer for the ODE function.
        (Default: 128)
    n_vae_hidden
        The dimensionality of the hidden layer for the AutoEncoder.
        (Default: 128)
    batch_norm
        Whether to include `BatchNorm` layer.
        (Default: 'False')
    ode_method
        ODE Solver method.
        (Default: 'euler')
    same_ode
        Whether to initialize all ODEs with same initialization
        (Default: 'False')
    """

    def __init__(
        self,
        n_int: int,
        n_latent: int = 20,
        n_sample: int = 1,
        n_ode_hidden: int = 128,
        n_vae_hidden: int = 128,
        batch_norm: bool = False,
        ode_method: str = 'dopri5',
        same_ode: bool = False,
        pre_ptime: bool = False,
        reconstruct_xt: bool = False,
    ):
        super().__init__()
        self.n_int = n_int
        self.n_latent = n_latent
        self.n_ode_hidden = n_ode_hidden
        self.n_vae_hidden = n_vae_hidden
        self.batch_norm = batch_norm
        self.ode_method = ode_method
        self.n_sample = n_sample
        self.same_ode = same_ode
        self.pre_ptime = pre_ptime
        self.reconstruct_xt = reconstruct_xt

        self.lode_func = nn.ModuleDict({f'node_{i}': LatentODE(n_latent, n_ode_hidden) for i in range(self.n_sample)})  
        self.encoder = Encoder(n_int, n_latent, n_vae_hidden, batch_norm)
        self.decoder = Decoder(n_int, n_latent, n_vae_hidden, batch_norm)
        self.beta, self.lam = torch.nn.Parameter(torch.Tensor(self.n_latent), requires_grad=True), torch.nn.Parameter(torch.Tensor(self.n_latent), requires_grad=True)
        nn.init.uniform_(self.beta)
        nn.init.uniform_(self.lam)
    def forward(self, s: torch.Tensor, u: torch.Tensor, Pt: torch.Tensor, g: torch.Tensor) -> tuple:
        """
        Given the transcriptomes and the treatments of cells, this function predicts the time, latent space and dynamics of the cells.

        Parameters
        ----------
        s
            Spliced reads.
        u
            Unspliced reads.
        Pt
            Pseudotime tensor (Only if you want your velocity to follow specific pseudotime)
        g
            Sample/Treatment index.

        Returns
        ----------
        5-tuple of :class:`torch.Tensor`
            Tensors for loss, including:
            1) total loss,
            2) reconstruction loss from encoder-derived latent space,
            3) reconstruction loss from ODE-solver latent space,
        """
        if self.pre_ptime:
            Ts, z_s = self.encoder(s)
            _, z_u = self.encoder(u)
            Ts = Ts.ravel()
            index = torch.argsort(Ts)
            Ts = Ts[index]
            s = s[index]
            z_s = z_s[index]
            u = u[index]
            z_u = z_u[index]
            g = g[index]
            Pt = Pt[index]

            index2 = (Ts[:-1] != Ts[1:])
            index2 = torch.cat((index2, torch.tensor([True]).to(index2.device))) ## index2 is used to get unique time points as odeint requires strictly increasing/decreasing time points
            Ts = Ts[index2]
            s = s[index2]
            z_s = z_s[index2]
            z_u = z_u[index2]
            u = u[index2]
            g = g[index2]
            Pt = Pt[index2]
            pt_loss = F.mse_loss(Pt, Ts, reduction='mean')
        else:
            Ts, z_s = self.encoder(s)
            _, z_u = self.encoder(u)
            Ts = Ts.ravel()
            index = torch.argsort(Ts)
            Ts = Ts[index]
            s = s[index]
            z_s = z_s[index]
            u = u[index]
            z_u = z_u[index]
            g = g[index]
            index2 = (Ts[:-1] != Ts[1:])
            index2 = torch.cat((index2, torch.tensor([True]).to(index2.device))) ## index2 is used to get unique time points as odeint requires strictly increasing/decreasing time points
            Ts = Ts[index2]
            s = s[index2]
            z_s = z_s[index2]

            z_u = z_u[index2]
            u = u[index2]
            g = g[index2]
            pt_loss = 0
        pred_z = torch.empty(s.size(0), self.n_latent).to(s.device)
        z_div = torch.Tensor([0]).to(s.device)
        for i in range(self.n_sample):
            mask = g == i
            
            if mask.sum() == 0:
                continue
                
            zsm = z_s[mask]
            zum = z_u[mask]
            Tm = Ts[mask]
            z0 = zsm[0]
            

            pred_z[mask] = odeint(self.lode_func[f'node_{i}'], z0, Tm, method = self.ode_method).view(-1, self.n_latent)
            z_div += F.mse_loss(self.lode_func[f'node_{i}'](Tm, pred_z[mask]), (torch.exp(self.beta)*zum-torch.exp(self.lam)*zsm), reduction='mean')
        pred_x_s = self.decoder(z_s)
        pred_x_u = self.decoder(z_u)
        if self.reconstruct_xt:
            pred_x_t = self.decoder(pred_z)
            recon_loss_xt = F.mse_loss(s, pred_x_t, reduction='mean')
        else:
            recon_loss_xt = F.mse_loss(z_s, pred_z, reduction='mean')
        recon_loss_ec = F.mse_loss(s, pred_x_s, reduction='mean')
        recon_loss_ec_u = F.mse_loss(u, pred_x_u, reduction='mean')

        loss =recon_loss_ec + recon_loss_ec_u + z_div + recon_loss_xt + pt_loss
        return loss, recon_loss_ec, recon_loss_ec_u, z_div
    



import torch
from torch.utils.data import DataLoader
from torchdiffeq import odeint
from typing import Optional, Union
from typing_extensions import Literal
import numpy as np
from anndata import AnnData
from scipy import sparse
from scipy.sparse import spmatrix
from scipy.sparse import issparse
from tqdm import tqdm
import os
from collections import defaultdict
from sklearn.decomposition import PCA
from .model import TNODE
from .data import split_data, MakeDataset



class Trainer:
    """
    Class for implementing the scTour training process.

    Parameters
    ----------
    adata
        An :class:`~anndata.AnnData` object.
    percent
        The percentage of cells used for model training. Default to 0.2 when the cell number > 10,000 and to 0.9 otherwise.
    n_latent
        The dimensionality of the latent space.
        (Default: 5)
    n_ode_hidden
        The dimensionality of the hidden layer for the latent ODE function.
        (Default: 25)
    n_vae_hidden
        The dimensionality of the hidden layer for the VAE.
        (Default: 128)
    batch_norm
        Whether to include a `BatchNorm` layer.
        (Default: `False`)
    odesample_obs
        key in anndata observation for ODE sample (This is different from celltype, use it only when you want to modelize different dynamics)
        (Default:None)
    ode_method
        The solver for integration.
        (Default: `'euler'`)
    nepoch
        Number of epochs.
    batch_size
        The batch size during training.
        (Default: 1024)
    lr
        The learning rate.
        (Default: 1e-3)
    wt_decay
        The weight decay for Adam optimizer.
        (Default: 1e-6)
    eps
        The eps for Adam optimizer.
        (Default: 0.01)
    random_state
        The seed for generating random numbers.
        (Default: 0)
    use_gpu
        Whether to use GPU when available
        (Default: True)
    layer
        RNA reads to use, either spliced or moments
        (Default: spliced)
    use_pca
        Whether to initialize the linear encoder with PCA
        (Default: False)
    reconstruct_xt
        Whether to recontsuct x(t) instead of z(t)
        (Default: False)
    """

    def __init__(
        self,
        adata: AnnData,
        percent: Optional[float] = None,
        odesample_obs: Optional[str] = None,
        n_latent: int = 5,
        n_ode_hidden: int = 25,
        n_vae_hidden: int = 128,
        batch_norm: bool = False,
        ode_method: str = 'dopri5',
        nepoch: Optional[int] = None,
        batch_size: int = 1024,
        lr: float = 1e-3,
        wt_decay: float = 1e-6,
        eps: float = 0.01,
        random_state: int = 0,
        use_gpu: bool = False,
        n_sample: int = 1,
        layer = 'spliced',
        use_pca: bool = False,
        pre_ptime: bool = False,
        reconstruct_xt: bool = False,
    ):
        self.adata = adata
        self.percent = percent
        self.precent = percent
        self.n_cells = adata.n_obs
        self.batch_size = batch_size
        self.nepoch = nepoch
        self.lr = lr
        self.wt_decay = wt_decay
        self.eps = eps
        self.n_latent = n_latent
        self.n_vae_hidden = n_vae_hidden 
        self.odesample_obs = odesample_obs
        self.pre_ptime = pre_ptime
        self.reconstruct_xt = reconstruct_xt
        
        if layer == 'spliced':
            self.layer = ['spliced','unspliced']
            print('Using spliced and unspliced reads')
        else:
            self.layer = ['Ms', 'Mu']
            print('Using spliced and unspliced moments')

        self.random_state = random_state
        np.random.seed(random_state)
        torch.manual_seed(random_state)
        
        #Convert sparse input to matrix
        if issparse(adata.layers[self.layer[0]]):
            self.adata.layers[self.layer[0]], self.adata.layers[self.layer[1]] = adata.layers[self.layer[0]].toarray(), adata.layers[self.layer[1]].toarray()

        
        if odesample_obs is None:
            self.adata.obs[odesample_obs] = 0
        assert self.adata.obs[odesample_obs].unique().size == n_sample
        self.n_int = adata.n_vars
        self.model_kwargs = dict(
            n_int = self.n_int,
            n_latent = self.n_latent,
            n_sample = n_sample,
            n_ode_hidden = n_ode_hidden,
            n_vae_hidden = n_vae_hidden,
            batch_norm = batch_norm,
            ode_method = ode_method,
            pre_ptime = pre_ptime,
            reconstruct_xt = reconstruct_xt
        )
        self.model = TNODE(**self.model_kwargs)
        self.log = defaultdict(list)
        
        if pre_ptime:
            ptime = adata.obs['latent_time']
        else:
            adata.obs['latent_time'] = np.ones(adata.n_obs)
        
        if use_pca:
            w1, w2 = self.get_init_pca()
            self.model.encoder.fc.L1.weight.data.copy_(w1)
            self.model.encoder.fc2.weight.data.copy_(w2)
            self.adata.layers[self.layer[0]] = self.adata.layers[self.layer[0]] - self.adata.layers[self.layer[0]].mean(0)
            self.adata.layers[self.layer[1]] = self.adata.layers[self.layer[1]] - self.adata.layers[self.layer[1]].mean(0)

        
        gpu = torch.cuda.is_available() and use_gpu
        if gpu:
            torch.cuda.manual_seed(random_state)
            self.device = torch.device('cuda')
        else:
            self.device = torch.device('cpu')
        self.model.to(self.device)
        
    def get_init_pca(self):
        pca1 = PCA(self.n_vae_hidden)
        pca2 = PCA(self.n_latent)
        reduced_x = pca1.fit_transform(self.adata.layers[self.layer[0]])
        pca2.fit(reduced_x)
        return torch.Tensor(pca1.components_), torch.Tensor(pca2.components_)
    
    def get_data_loaders(self) -> None:
        """
        Generate Data Loaders for training and validation datasets.
        """

        train_data, val_data = split_data(self.adata, self.percent,self.odesample_obs, self.layer)
        self.train_dataset = MakeDataset(train_data)
        self.val_dataset = MakeDataset(val_data)

        self.train_dl = DataLoader(self.train_dataset, batch_size = self.batch_size, shuffle = True)
        self.val_dl = DataLoader(self.val_dataset, batch_size = self.batch_size)


    def train(self):
        self.get_data_loaders()
        params = filter(lambda p: p.requires_grad, self.model.parameters())
        self.optimizer = torch.optim.Adam(params, lr = self.lr, weight_decay = self.wt_decay, eps = self.eps)
        #self.optimizer = torch.optim.SGD(params, lr=self.lr, weight_decay=self.wt_decay, momentum=0.9)
        with tqdm(total=self.nepoch, unit='epoch') as t:
            for tepoch in range(t.total):
                train_loss = self.on_epoch_train(self.train_dl)
                val_loss = self.on_epoch_val(self.val_dl)
                self.log['train_loss'].append(train_loss.item())
                self.log['validation_loss'].append(val_loss.item())
                t.set_description(f"Epoch {tepoch + 1}")
                t.set_postfix({'train_loss': train_loss.item(), 'val_loss':val_loss.item()}, refresh=False)
                t.update()


    def on_epoch_train(self, DL) -> float:
        """
        Go through the model and update the model parameters.

        Parameters
        ----------
        DL
            DataLoader for training dataset.

        Returns
        ----------
        float
            Training loss for the current epoch.
        """
        self.model.train()
        total_loss = .0
        ss = 0
        for X, Y, T, S in DL:
            self.optimizer.zero_grad()
            X = X.to(self.device)
            Y = Y.to(self.device)
            T = T.to(self.device)
            S = S.to(self.device)
            loss, _, _, _ = self.model(X,Y,T,S)
            loss.backward()
            self.optimizer.step()

            total_loss += loss * X.size(0)
            ss += X.size(0)

        train_loss = total_loss/ss
        return train_loss


    @torch.no_grad()
    def on_epoch_val(self, DL) -> float:
        """
        Validate using validation dataset.

        Parameters
        ----------
        DL
            DataLoader for validation dataset.

        Returns
        ----------
        float
            Validation loss for the current epoch.
        """

        self.model.eval()
        total_loss = .0
        ss = 0
        for X, Y, T, S in DL:
            X = X.to(self.device)
            Y = Y.to(self.device)
            T = T.to(self.device)
            S = S.to(self.device)
            loss, _,_,_ = self.model(X, Y, T, S)
            total_loss += loss * X.size(0)
            ss += X.size(0)

        val_loss = total_loss/ss
        return val_loss

    def save_model(
        self,
        save_dir: str,
        save_prefix: str,
    ) -> None:
        """
        Save the model.

        Parameters
        ----------
        save_dir
            The directory where the model is saved.
        save_prefix
            The prefix for model name.
        """

        save_path = os.path.abspath(os.path.join(save_dir, f'{save_prefix}.pth'))
        torch.save(
            {
                'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'var_names': self.adata.var_names,
                'nepoch': self.nepoch,
                'random_state': self.random_state,
                'percent': self.percent,
                'model_kwargs': self.model_kwargs,
                'log': self.log,
            },
            save_path
        )



import numpy as np
from scipy.sparse import issparse
from .model import TNODE
from torch.autograd.functional import jacobian
import torch
from sklearn.metrics.pairwise import cosine_similarity
import copy
import anndata as ad
import pandas as pd




#Class to perform gene ranking. Output can be used directly in gseapy for pathways
class ModelAnalyzer:
    def __init__(self, data, n_vectors, models_path, layer='spliced', cops=None, label='clusters'):
        self.data = data
        self.n_vectors = n_vectors
        self.models_path = models_path
        self.layer = layer
        self.cops = cops
        self.label = label
    def post_training_enc(self):
        with torch.no_grad():
            if issparse(self.data.layers[self.layer]):
                self.data.obs['ptime'], self.data.obsm['X_z'] = self.model.encoder(
                    torch.Tensor(self.data.layers[self.layer].toarray()))
            else:
                self.data.obs['ptime'], self.data.obsm['X_z'] = self.model.encoder(
                    torch.Tensor(self.data.layers[self.layer]))
            self.data.obsm['X_z'] = self.data.obsm['X_z'].numpy()
            return self.data

    def load_model(self, model_path):
        model_dict = torch.load(model_path, map_location=torch.device('cpu'))
        model_kwargs = model_dict['model_kwargs']
        self.model = TNODE(**model_kwargs)
        self.model.load_state_dict(model_dict['model_state_dict'])
        return self.model

    def compute_jacobian_eigenvalues(self):
        eigen_values = {}
        gene_dic = {}
        gene_dic_eig = {}
        self.data = self.post_training_enc()
        if self.model.n_sample == 1:
            masked = self.data
            if self.cops is None:
                cop = np.mean(masked.obsm['X_z'], axis=0)
            else:
                cop = np.mean(masked[masked.obs[self.label]==self.cops[0]].obsm['X_z'], axis=0)

            self.model.eval()
            jac0 = jacobian(self.model.lode_func[f'node_0'], (torch.Tensor([1]), torch.Tensor(cop)))[1]

            val, vec = np.linalg.eig(jac0)
            _, sorted_indices = np.unique(-np.absolute(val), return_index=True)

            sorted_indices = sorted_indices[:self.n_vectors]

            eigen_values[masked.obs['treatment'][0]] = val[sorted_indices]
            decoded_v = self.model.decoder(torch.Tensor(vec[:, sorted_indices].T)).detach().numpy()
            indic = np.argsort(-np.abs(decoded_v))

            gene_dic_eig[masked.obs['treatment'][0]] = decoded_v
            gene_dic[masked.obs['treatment'][0]] = self.data.var.index.to_numpy()[indic]
        else:
            for j in range(self.model.n_sample):
                mask = self.data.obs['sample'] == j
                masked = self.data[mask]
                cop = np.mean(masked.obsm['X_z'], axis=0)
    
                self.model.eval()
                jac0 = jacobian(self.model.lode_func[f'node_{j}'], (torch.Tensor([1]), torch.Tensor(cop)))[1]
    
                val, vec = np.linalg.eig(jac0)
                _, sorted_indices = np.unique(-np.absolute(val), return_index=True)
    
                sorted_indices = sorted_indices[:self.n_vectors]
    
                eigen_values[masked.obs['treatment'][0]] = val[sorted_indices]
                decoded_v = self.model.decoder(torch.Tensor(vec[:, sorted_indices].T)).detach().numpy()
                indic = np.argsort(-np.abs(decoded_v))
    
                gene_dic_eig[masked.obs['treatment'][0]] = decoded_v
                gene_dic[masked.obs['treatment'][0]] = self.data.var.index.to_numpy()[indic]

        return eigen_values, gene_dic, gene_dic_eig

    def models_output(self):
        self.results = {}

        for o, model_path in enumerate(self.models_path):
            model_key = f'model {o}'
            self.model = self.load_model(model_path)

            eigen_values, genes, eigen_vectors = self.compute_jacobian_eigenvalues()

            self.results[model_key] = {
                'eigen_values': eigen_values,
                'genes': genes,
                'eigen_vectors': eigen_vectors
            }

        return self.results

    def calculate_eigenvector_similarity(self, treatment):
        sim = None
        for n, model_a in enumerate(self.results.values()):
            a = model_a['eigen_vectors'][treatment]
            for m, model_b in enumerate(self.results.values()):
                if n != m:
                    b = model_b['eigen_vectors'][treatment]
                    cos_matrix = cosine_similarity(a, b)

                    eigen_sim = np.array(
                        [np.ones(self.n_vectors, dtype=int) * n, np.ones(self.n_vectors, dtype=int) * m,
                         np.arange(self.n_vectors, dtype=int), np.argmax(np.abs(cosine_similarity(a, b)), axis=1)]).T
                    if sim is None:
                        sim = eigen_sim
                    else:
                        sim = np.append(sim, eigen_sim, axis=0)
        self.similarity_matrix = sim
        return sim


    def gene_order(self, gene_df):
        self.new_df = gene_df.copy()
        for column in self.new_df.columns:
            self.new_df.loc[:, column] = self.new_df.index[self.new_df.loc[:, column].argsort()]
        self.new_df.index = np.arange(self.new_df.shape[0])+1
        return self.new_df

    def gene_ranking(self):
        all_gene_rank, all_gene_mean = {}, {} #Dictionary to save all information about genes ranking accross treatment and eigenvector
        gene_eigen_treatment_mean = {} #The output dictionary. It contains DataFrame of each treatment with genes average rank for each eigenvector.
        gene_eigen_treatment_order = {}
        for t in self.data.obs['treatment'].unique():
            sim = self.calculate_eigenvector_similarity(t)
            treatment = t
            model = 0
            eigens = np.arange(self.n_vectors)
            all_gene_rank[treatment] = {}
            all_gene_mean[treatment] = {}
            for eigen in eigens:
                mask = (self.similarity_matrix[:, 0].astype(int) == model) & (
                            self.similarity_matrix[:, 2].astype(int) == eigen)
                new_ev = self.similarity_matrix[mask]

                l = self.results[f'model {model}']['genes'][treatment][eigen]
                for i in range(new_ev.shape[0]):
                    l = np.vstack((l, self.results[f'model {new_ev[i, 1]}']['genes'][treatment][new_ev[i, 3]]))

                genes_rank = {}
                mean_gene = {}
                for k in self.data.var.index:
                    genes_rank[k] = np.where(l == k)[1] + 1
                    mean_gene[k] = np.mean(genes_rank[k])

                all_gene_rank[treatment][eigen] = genes_rank
                all_gene_mean[treatment][eigen] = mean_gene
            
            gene_eigen_treatment_mean[treatment] = pd.DataFrame(all_gene_mean[treatment])
            gene_eigen_treatment_order[treatment] = self.gene_order(gene_eigen_treatment_mean[treatment])
            
        self.all_gene_rank = all_gene_rank
        return gene_eigen_treatment_order,gene_eigen_treatment_mean


def vector_fields_similarity(adata, models, layer="spliced"):
    #Iterate over all models and return velocity_field
    matrix_dict = {}
    for m in models:
        matrix_dict[m] = decode_gene_velocity(adata, m, layer)
        
    # Initialize an empty dictionary to store cosine similarity values
    cos_sim_dict = {}
    # Initialize a set to keep track of processed pairs
    processed_pairs = set()
    # Iterate over all pairs of matrices
    for key1, matrix1 in matrix_dict.items():
        for key2, matrix2 in matrix_dict.items():
            if key1 != key2 and (key1, key2) not in processed_pairs and (key2, key1) not in processed_pairs:
                # Compute cosine similarity between the 2 matrices
                cos_sim = np.diag(cosine_similarity(matrix1, matrix2))
                # Store the cosine similarity value in the dictionary
                cos_sim_dict[(key1, key2)] = np.abs(np.mean(cos_sim))
                # Mark this pair as processed
                processed_pairs.add((key1, key2))
    return cos_sim_dict, np.mean(list(cos_sim_dict.values()))


def decode_gene_velocity(adata,model, layer='spliced'):
    model_dic = torch.load(model)
    kwargs = model_dic['model_kwargs']
    trained_model = TNODE(**kwargs)
    trained_model.load_state_dict(model_dic['model_state_dict'])
    trained_model.eval()
    if issparse(adata.layers[layer]):
        encoded = trained_model.encoder(torch.Tensor(adata.layers[layer].toarray()))
    else:
        encoded = trained_model.encoder(torch.Tensor(adata.layers[layer]))

    ptime , z_s = encoded[0].detach().numpy(), encoded[1].detach().numpy()
    vf = np.empty(encoded[1].shape)
    for j in adata.obs['sample'].unique():
        mask = adata.obs['sample'] == j
        vf[mask] = trained_model.lode_func[f'node_{j}'](encoded[0][mask].values, encoded[1][mask]).detach().numpy()
    return trained_model.decoder(torch.Tensor(vf)).cpu().detach().numpy()

def latent_data(adata, model_pth,layer='spliced'):
    model_dic = torch.load(model_pth,map_location=torch.device('cpu'))
    kwargs = model_dic['model_kwargs']
    trained_model = TNODE(**kwargs)
    trained_model.load_state_dict(model_dic['model_state_dict'])
    trained_model.eval()

    if issparse(adata.layers[layer]):
        encoded = trained_model.encoder(torch.Tensor(adata.layers[layer].toarray()))
    else:
        encoded = trained_model.encoder(torch.Tensor(adata.layers[layer]))

    ptime , z_s = encoded[0].detach().numpy(), encoded[1].detach().numpy()
    vf = np.empty(encoded[1].shape)
    for j in adata.obs['sample'].unique():
        mask = adata.obs['sample'] == j
        vf[mask] = trained_model.lode_func[f'node_{j}'](encoded[0][mask].values, encoded[1][mask]).detach().numpy()
    
    adata.obs['ptime'] = ptime
    latent_adata = ad.AnnData(z_s, 
                          obsm={'X_z': z_s},
                         layers = {'spliced': z_s,
                                   'spliced_velocity': vf},
                          obs=adata.obs)
    del adata.obs['ptime']
    return latent_adata



## All these function are the same of UniTVelo

#%%
"""
Evaluation utility functions.
This module contains util functions for computing evaluation scores.
"""

def summary_scores(all_scores):
    """Summarize group scores.
    
    Args:
        all_scores (dict{str,list}): 
            {group name: score list of individual cells}.
    
    Returns:
        dict{str,float}: 
            Group-wise aggregation scores.
        float: 
            score aggregated on all samples
        
    """
    sep_scores = {k:np.mean(s) for k, s in all_scores.items() if s}
    overal_agg = np.mean([s for k, s in sep_scores.items() if s])
    return sep_scores, overal_agg

def keep_type(adata, nodes, target, k_cluster):
    """Select cells of targeted type
    
    Args:
        adata (Anndata): 
            Anndata object.
        nodes (list): 
            Indexes for cells
        target (str): 
            Cluster name.
        k_cluster (str): 
            Cluster key in adata.obs dataframe

    Returns:
        list: 
            Selected cells.

    """
    return nodes[adata.obs[k_cluster][nodes].values == target]

def cross_boundary_correctness(
    adata, 
    k_cluster, 
    k_velocity, 
    cluster_edges, 
    return_raw=False, 
    x_emb="X_umap"
):
    """Cross-Boundary Direction Correctness Score (A->B)
    
    Args:
        adata (Anndata): 
            Anndata object.
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        cluster_edges (list of tuples("A", "B")): 
            pairs of clusters has transition direction A->B
        return_raw (bool): 
            return aggregated or raw scores.
        x_emb (str): 
            key to x embedding for visualization.
        
    Returns:
        dict: 
            all_scores indexed by cluster_edges or mean scores indexed by cluster_edges
        float: 
            averaged score over all cells.
        
    """
    scores = {}
    all_scores = {}
    
    if x_emb == "X_umap":
        v_emb = adata.obsm['{}_umap'.format(k_velocity)]
    else:
        v_emb = adata.obsm[[key for key in adata.obsm if key.startswith(k_velocity)][0]]
        
    x_emb = adata.obsm[x_emb]
    
    for u, v in cluster_edges:
        sel = adata.obs[k_cluster] == u
        nbs = adata.uns['neighbors']['indices'][sel] # [n * 30]
        
        boundary_nodes = map(lambda nodes:keep_type(adata, nodes, v, k_cluster), nbs)
        x_points = x_emb[sel]
        x_velocities = v_emb[sel]
        
        type_score = []
        for x_pos, x_vel, nodes in zip(x_points, x_velocities, boundary_nodes):
            if len(nodes) == 0: continue

            position_dif = x_emb[nodes] - x_pos
            dir_scores = cosine_similarity(position_dif, x_vel.reshape(1,-1)).flatten()
            type_score.append(np.mean(dir_scores))
        
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
        
    if return_raw:
        return all_scores 
    
    return scores, np.mean([sc for sc in scores.values()])

def inner_cluster_coh(adata, k_cluster, k_velocity, return_raw=False):
    """In-cluster Coherence Score.
    
    Args:
        adata (Anndata): 
            Anndata object.
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        return_raw (bool): 
            return aggregated or raw scores.
        
    Returns:
        dict: 
            all_scores indexed by cluster_edges mean scores indexed by cluster_edges
        float: 
            averaged score over all cells.
        
    """
    clusters = np.unique(adata.obs[k_cluster])
    scores = {}
    all_scores = {}

    for cat in clusters:
        sel = adata.obs[k_cluster] == cat
        nbs = adata.uns['neighbors']['indices'][sel]
        same_cat_nodes = map(lambda nodes:keep_type(adata, nodes, cat, k_cluster), nbs)

        velocities = adata.layers[k_velocity]
        cat_vels = velocities[sel]
        cat_score = [cosine_similarity(cat_vels[[ith]], velocities[nodes]).mean() 
                     for ith, nodes in enumerate(same_cat_nodes) 
                     if len(nodes) > 0]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)
    
    if return_raw:
        return all_scores
    
    return scores, np.mean([sc for sc in scores.values()])

def evaluate(
    adata, 
    cluster_edges, 
    k_cluster, 
    k_velocity="velocity", 
    x_emb="X_umap", 
    verbose=True
):
    """Evaluate velocity estimation results using 5 metrics.
    
    Args:
        adata (Anndata): 
            Anndata object.
        cluster_edges (list of tuples("A", "B")): 
            pairs of clusters has transition direction A->B
        k_cluster (str): 
            key to the cluster column in adata.obs DataFrame.
        k_velocity (str): 
            key to the velocity matrix in adata.obsm.
        x_emb (str): 
            key to x embedding for visualization.
        
    Returns:
        dict: 
            aggregated metric scores.
    
    """
    crs_bdr_crc = cross_boundary_correctness(adata, k_cluster, k_velocity, cluster_edges, True, x_emb)
    ic_coh = inner_cluster_coh(adata, k_cluster, k_velocity, True)
    
    if verbose:
        print("# Cross-Boundary Direction Correctness (A->B)\n{}\nTotal Mean: {}".format(*summary_scores(crs_bdr_crc)))
        print("# In-cluster Coherence\n{}\nTotal Mean: {}".format(*summary_scores(ic_coh)))
    
    return {
        "Cross-Boundary Direction Correctness (A->B)": summary_scores(crs_bdr_crc),
        "In-cluster Coherence": summary_scores(ic_coh),
    }


import torch
import torch.nn as nn


class LatentODE(nn.Module):
    """
    A class modelling the latent splicing dynamics.

    Parameters
    ----------
    n_latent
        Dimension of latent space.
        (Default: 20)
    n_hidden
        The dimensionality of the hidden layer for the ODE function.
        (Default: 128)
    """

    def __init__(
        self,
        n_latent: int = 20,
        n_hidden: int = 128,
    ):
        super().__init__()
        self.elu = nn.ELU()
        self.fc1 = nn.Linear(n_latent, n_hidden)
        self.fc2 = nn.Linear(n_hidden, n_latent)

    def forward(self, t: torch.Tensor, x: torch.Tensor):
        """
        Compute the gradient at a given time t and a given state x.

        Parameters
        ----------
        t
            A given time point.
        x
            A given spliced latent state.

        Returns
        ----------
        :class:`torch.Tensor`
            A tensor
        """
        out = self.fc1(x)
        out = self.elu(out)
        out = self.fc2(out)
        return out


class Encoder(nn.Module):
    """
    Linear encoder class for dimensionality reduction and features selection.

    Parameters
    ----------
    n_int
        Number of genes.
    n_latent
        Dimension of latent space.
        (Default: 20)
    n_hidden
        The dimensionality of the hidden layer for the AutoEncoder.
        (Default: 128)
    batch_norm
        Whether to include `BatchNorm` layer or not.
        (Default: 'False')
    """

    def __init__(
        self,
        n_int: int,
        n_latent: int = 20,
        n_hidden: int = 128,
        batch_norm: bool = False,
    ):
        super().__init__()
        
        self.n_latent = n_latent
        self.fc = nn.Sequential()
        self.fc.add_module('L1', nn.Linear(n_int, n_hidden))
        if batch_norm:
            self.fc.add_module('N1', nn.BatchNorm1d(n_hidden))
        self.fc2 = nn.Linear(n_hidden, n_latent)
        self.fc3 = nn.Linear(n_hidden, 1)
        

    def forward(self, x:torch.Tensor) -> tuple:
        x = self.fc(x)
        out = self.fc2(x)
        t = self.fc3(x).sigmoid()
        return t, out


class Decoder(nn.Module):
    """
    Linear decoder class to reconstruct the original counts based on the latent representation.

    Parameters
    ----------
    n_latent
        Dimension of latent space.
        (Default: 20)
    n_int
        Number of genes.
    n_hidden
        The dimensionality of the hidden layer for the AutoEncoder.
        (Default: 128)
    batch_norm
        Whether to include `BatchNorm` layer or not.
        (Default: 'False')
    """

    def __init__(
        self,
        n_int: int,
        n_latent: int = 20,
        n_hidden: int = 128,
        batch_norm: bool = False,
    ):
        super().__init__()
        self.fc = nn.Sequential()
        self.fc.add_module('L1', nn.Linear(n_latent, n_hidden))
        if batch_norm:
            self.fc.add_module('N1', nn.BatchNorm1d(n_hidden))
        self.fc2 = nn.Linear(n_hidden, n_int)

    def forward(self, z: torch.Tensor):
        out = self.fc(z)
        recon_x = self.fc2(out)
        return recon_x



import numpy as np
import torch
from scipy.sparse import issparse
from torch.autograd.functional import jacobian
import networkx as nx
import matplotlib.pyplot as plt
from .model import TNODE
from .utils import latent_data
import pandas as pd
from matplotlib.lines import Line2D
import scvelo as scv
from matplotlib.animation import FuncAnimation, PillowWriter

class GraphMaker:
    def __init__(self, adata, model_path, layer='spliced'):
        self.adata = adata
        self.model_path = model_path
        self.layer = layer

    def assign_samples(self, adata, n_sam=1):
        # Assuming this function assigns samples to adata
        adata.obs['sample'] = 0
        return adata

    def graph_maker(self, g, n_frames=30, celltype_to_linearize=None, take_windows=False, ci=0, reverse=False):
        s = 0
        gi = [np.where(self.adata.var.index == gene)[0][0] for gene in g if gene in self.adata.var.index]
        self.adata = self.assign_samples(self.adata, n_sam=1)

        model_dict = torch.load(self.model_path, map_location=torch.device('cpu'))
        model_kwargs = model_dict['model_kwargs']
        model = TNODE(**model_kwargs)
        model.load_state_dict(model_dict['model_state_dict'])

        latent_adata = latent_data(self.adata, self.model_path, layer=self.layer)
        if reverse:
            latent_adata.obs['ptime'] = 1 - latent_adata.obs['ptime']
        window_size = int(latent_adata.n_obs / n_frames)
        cell_of_ptime, cop = self._get_cop(latent_adata, celltype_to_linearize, take_windows, ci, window_size)

        model.eval()
        jac0 = jacobian(model.lode_func[f'node_{s}'], (torch.Tensor([1]), torch.Tensor(cop)))[1]
        combined_weights, _ = self.get_wab(model)
        combined_weights_enc, _ = self.get_wab_encoder(model)
        A = combined_weights @ (jac0.numpy()) @ combined_weights_enc
        A = A.T #so the graph is inherently from i to j -opposite to what written in the paper
        cell_dict = latent_adata[cell_of_ptime].obs['clusters'].value_counts().to_dict()
        sorted_cell_value = {k: v for k, v in sorted(cell_dict.items(), key=lambda item: item[1], reverse=True)}
        return A, gi, g, np.mean(latent_adata[cell_of_ptime].obs['ptime']), sorted_cell_value, cell_of_ptime

    def _get_cop(self, latent_adata, celltype_to_linearize, take_windows, ci, window_size):
        if take_windows and celltype_to_linearize is None:
            cell_of_ptime = latent_adata.obs['ptime'].sort_values().index[ci:ci+window_size]
        elif take_windows and celltype_to_linearize is not None:
            cell_of_ptime = latent_adata[latent_adata.obs['clusters'].isin(celltype_to_linearize)].obs['ptime'].sort_values().index[ci:ci+window_size]
        elif celltype_to_linearize is not None:
            cell_of_ptime = latent_adata[latent_adata.obs['clusters'].isin(celltype_to_linearize)].index
        else:
            cell_of_ptime = latent_adata.obs['ptime'].index

        return cell_of_ptime,np.mean(latent_adata[cell_of_ptime].obsm['X_z'], axis=0)

    def create_graph(self, A, gi, g, threshold_a, threshold_b=None):
        heatmap_data = self.give_small(A, gi).numpy()
        if threshold_b is None:
            significant_connections = np.where(abs(heatmap_data) > threshold_a)
        else:
            significant_connections = np.where((abs(heatmap_data) > threshold_a) & (abs(heatmap_data) < threshold_b))

        G = nx.MultiDiGraph()
        num_nodes = heatmap_data.shape[0]
        for i in range(num_nodes):
            G.add_node(g[i])

        for i, j in zip(significant_connections[0], significant_connections[1]):
            if i != j:
                weight = heatmap_data[i, j]
                G.add_edge(g[i], g[j], weight=weight, direction='forward')

        return G

    def give_small(self, matrix, gi):
        return matrix[gi][:, gi].squeeze()

    def get_wab(self, model):
        weights = []
        for name, param in model.decoder.named_parameters():
            if 'weight' in name or 'bias' in name:
                weights.append(param.data)

        combined_weights = torch.matmul(weights[2], weights[0])
        combined_bias = torch.matmul(weights[2], weights[1]) + weights[3]
        return combined_weights, combined_bias

    def get_wab_encoder(self, model):
        weights = []
        for name, param in model.encoder.named_parameters():
            if 'weight' in name and 'fc3' not in name:
                weights.append(param.data)
            elif 'bias' in name and 'fc3' not in name:
                weights.append(param.data)

        combined_weights = torch.mm(weights[2], weights[0])
        combined_bias = torch.mv(weights[2], weights[1]) + weights[3]
        return combined_weights, combined_bias

    def plot_custom_graph(self, G, node_color='#EAE0D5', positive_color='#A3320B', negative_color='#61988E', f_color='black', edge_width=1,ax_network=None,network_node_file=None):
        if ax_network is None:
            fig, ax_network = plt.subplots(figsize=(12, 8))
        
        pos = nx.spring_layout(G)  # Initial position layout
        if network_node_file is not None:
            network_node = pd.read_csv(network_node_file, delimiter='\t', index_col=0)
            # Filter out nodes not present in the network_node file
            nodes_to_remove = [node for node in pos.keys() if node not in network_node.index]
            G.remove_nodes_from(nodes_to_remove)
            pos = {node: pos[node] for node in pos if node in network_node.index}
            print("The following genes are either not protein coding genes or not found in string-db: ", nodes_to_remove)
            for i in pos.keys():
                pos[i][0] = network_node.loc[i, 'x_position']
                pos[i][1] = 1 - network_node.loc[i, 'y_position']
                
        nx.draw_networkx_nodes(G, pos, ax=ax_network, node_color=node_color, node_size=1500)
        nx.draw_networkx_labels(G, pos, ax=ax_network, font_family='DejaVu Sans', font_size=10, font_color=f_color)

        for u, v, data in G.edges(data=True):
            color = positive_color if data['weight'] > 0 else negative_color
            width = abs(data['weight']) * edge_width
            nx.draw_networkx_edges(G, pos, ax=ax_network, edgelist=[(u, v)], width=width, edge_color=color, arrows=True, arrowsize=10, connectionstyle=f"arc3,rad={0.10}", node_size=1500)
            
        # Create custom legend
        legend_elements = [
            Line2D([0], [0], color=positive_color, lw=4, label='Positive'),
            Line2D([0], [0], color=negative_color, lw=4, label='Negative')
        ]
        ax_network.legend(handles=legend_elements, loc='best')
        
        plt.title('Gene Interaction Network')
        plt.show()



class GraphMakerAnimation:
    def __init__(self, adata, model_path,cell_label, layer='spliced', reverse=False, show_scatter=False):
        self.adata = adata
        self.model_path = model_path
        self.layer = layer
        self.adata_subset = None 
        self.fig = None
        self.ax_network = None
        self.ax_scatter = None
        self.animation = None
        self.celltype_to_linearize = None
        self.reverse = reverse
        self.fix_pos = None
        self.cell_label = cell_label
        self.show_scatter = show_scatter
    def assign_samples(self, adata, n_sam=1):
        adata.obs['sample'] = 0
        return adata

    def graph_maker(self, g, celltype_to_linearize=None, take_windows=False, ci=0, reverse=False):
        s = 0
        gi = [np.where(self.adata.var.index == gene)[0][0] for gene in g if gene in self.adata.var.index]
        self.adata = self.assign_samples(self.adata, n_sam=1)

        model_dict = torch.load(self.model_path, map_location=torch.device('cpu'))
        model_kwargs = model_dict['model_kwargs']
        model = TNODE(**model_kwargs)
        model.load_state_dict(model_dict['model_state_dict'])

        latent_adata = latent_data(self.adata, self.model_path, layer=self.layer)
        if reverse:
            latent_adata.obs['ptime'] = 1 - latent_adata.obs['ptime']

        if self.celltype_to_linearize is not None:
            latent_adata = latent_adata[latent_adata.obs[self.cell_label].isin(self.celltype_to_linearize)]

        self.sorted_cell_index = latent_adata.obs['ptime'].sort_values().index.tolist()

        self.window_size = int(len(self.sorted_cell_index) / self.n_frames)
        cell_of_ptime, cop = self._get_cop(latent_adata, celltype_to_linearize, take_windows, ci, self.window_size)

        model.eval()
        jac0 = jacobian(model.lode_func[f'node_{s}'], (torch.Tensor([1]), torch.Tensor(cop)))[1]
        combined_weights, _ = self.get_wab(model)
        combined_weights_enc, _ = self.get_wab_encoder(model)
        A = combined_weights @ (jac0.numpy()) @ combined_weights_enc
        A = A.T #so the graph is inherently from i to j -opposite to what written in the paper
        cell_dict = latent_adata[cell_of_ptime].obs[self.cell_label].value_counts().to_dict()
        sorted_cell_value = {k: v for k, v in sorted(cell_dict.items(), key=lambda item: item[1], reverse=True)}
        return A, gi, g, np.mean(latent_adata[cell_of_ptime].obs['ptime']), sorted_cell_value, cell_of_ptime

    def _get_cop(self, latent_adata, celltype_to_linearize, take_windows, ci, window_size):
        if take_windows and celltype_to_linearize is None:
            cell_of_ptime = latent_adata.obs['ptime'].sort_values().index[ci:ci+window_size]
        elif take_windows and celltype_to_linearize is not None:
            cell_of_ptime = latent_adata[latent_adata.obs[self.cell_label].isin(celltype_to_linearize)].obs['ptime'].sort_values().index[ci:ci+window_size]
        elif celltype_to_linearize is not None:
            cell_of_ptime = latent_adata[latent_adata.obs[self.cell_label].isin(celltype_to_linearize)].index
        else:
            cell_of_ptime = latent_adata.obs['ptime'].index
        return cell_of_ptime, np.mean(latent_adata[cell_of_ptime].obsm['X_z'], axis=0)

    def create_graph(self, A, gi, g, threshold_a, threshold_b=None):
        heatmap_data = self.give_small(A, gi).numpy()
        if threshold_b is None:
            significant_connections = np.where(abs(heatmap_data) > threshold_a)
        else:
            significant_connections = np.where((abs(heatmap_data) > threshold_a) & (abs(heatmap_data) < threshold_b))

        G = nx.MultiDiGraph()
        num_nodes = heatmap_data.shape[0]
        for i in range(num_nodes):
            G.add_node(g[i])

        for i, j in zip(significant_connections[0], significant_connections[1]):
            if i != j:
                weight = heatmap_data[i, j]
                G.add_edge(g[i], g[j], weight=weight, direction='forward')
        return G

    def give_small(self, matrix, gi):
        return matrix[gi][:, gi].squeeze()

    def get_wab(self, model):
        weights = []
        for name, param in model.decoder.named_parameters():
            if 'weight' in name or 'bias' in name:
                weights.append(param.data)

        combined_weights = torch.matmul(weights[2], weights[0])
        combined_bias = torch.matmul(weights[2], weights[1]) + weights[3]
        return combined_weights, combined_bias

    def get_wab_encoder(self, model):
        weights = []
        for name, param in model.encoder.named_parameters():
            if 'weight' in name and 'fc3' not in name:
                weights.append(param.data)
            elif 'bias' in name and 'fc3' not in name:
                weights.append(param.data)

        combined_weights = torch.mm(weights[2], weights[0])
        combined_bias = torch.mv(weights[2], weights[1]) + weights[3]
        return combined_weights, combined_bias

    def plot_custom_graph(self, G, node_color='#EAE0D5', positive_color='#A3320B', negative_color='#61988E', 
                         f_color='black', edge_width=1, ax_network=None, network_node_file=None):
        if ax_network is None:
            fig, ax_network = plt.subplots(figsize=(12, 8))
        
        pos = nx.spring_layout(G, seed=42)
        if network_node_file is not None:
            network_node = pd.read_csv(network_node_file, delimiter='\t', index_col=0)
            nodes_to_remove = [node for node in pos.keys() if node not in network_node.index]
            G.remove_nodes_from(nodes_to_remove)
            pos = {node: pos[node] for node in pos if node in network_node.index}
            print("Removed genes:", nodes_to_remove)
            for i in pos.keys():
                pos[i][0] = network_node.loc[i, 'x_position']
                pos[i][1] = 1 - network_node.loc[i, 'y_position']
                
        nx.draw_networkx_nodes(G, pos, ax=ax_network, node_color=node_color, node_size=1500)
        nx.draw_networkx_labels(G, pos, ax=ax_network, font_family='DejaVu Sans', font_size=10, font_color=f_color)

        for u, v, data in G.edges(data=True):
            color = positive_color if data['weight'] > 0 else negative_color
            width = abs(data['weight']) * edge_width
            nx.draw_networkx_edges(G, pos, ax=ax_network, edgelist=[(u, v)], width=width, edge_color=color, 
                                   arrows=True, arrowsize=10, connectionstyle=f"arc3,rad={0.10}", node_size=1500)
            
        legend_elements = [
            Line2D([0], [0], color=positive_color, lw=4, label='Positive'),
            Line2D([0], [0], color=negative_color, lw=4, label='Negative')
        ]
        ax_network.legend(handles=legend_elements, loc='best')
        plt.title('Gene Interaction Network')
        plt.show()

    def create_animation(self, g, path_to_save ,adata_subset=None, n_frames=20, threshold_a=0.01, threshold_b=None, 
                        network_node_file=None, interval=100, celltype_to_linearize = None):
        if adata_subset is None:
            self.adata_subset = self.adata.copy()
        else:
            self.adata_subset = adata_subset.copy()
        self.celltype_to_linearize = celltype_to_linearize
        self.g = g
        self.n_frames = n_frames
        self.threshold_a = threshold_a
        self.threshold_b = threshold_b
        self.network_node_file = network_node_file


        self.graph_maker(g, self.celltype_to_linearize, take_windows=True)

        self.fig, (self.ax_network, self.ax_scatter) = plt.subplots(1, 2, figsize=(25, 10))

        frames = range(0, len(self.sorted_cell_index), self.window_size)
        
        valid_frames = (frame for frame in frames if self._is_valid_frame(frame))
        self.animation = FuncAnimation(self.fig, self._update, frames=valid_frames, 
                                      interval=interval, blit=False)
        self.animation.save(path_to_save, writer=PillowWriter(fps=2))
        plt.show()

    def _is_valid_frame(self, start_idx):
        window = self.sorted_cell_index[start_idx : start_idx + self.window_size]
        return len(window) > 0
    
    def _is_all_na(self, cell_n):
        self.adata_subset.obs['fixed_clusters'] = self.adata_subset.obs[self.cell_label]
        self.adata_subset.obs['changed_clusters'] = pd.Categorical(
            [pd.NA] * len(self.adata_subset),
            categories=self.adata_subset.obs['fixed_clusters'].cat.categories
        )

        cell_index = self.adata_subset.obs_names[cell_n:cell_n + self.n_frames]
        for cell in cell_index:
            if cell in self.adata_subset.obs_names:
                idx = self.adata_subset.obs_names.get_loc(cell)
                cluster = self.adata_subset.obs['fixed_clusters'][idx]
                self.adata_subset.obs.at[cell, 'changed_clusters'] = cluster

        return self.adata_subset.obs['changed_clusters'].isna().all()

    
    def _update(self, cell_n):
        self.ax_network.clear()
        self.ax_network.axis('off')
        self.ax_scatter.clear()
        # Get network data
        A, gi, _, aptime, _, cell_index = self.graph_maker(
            self.g, take_windows=True, ci=cell_n, celltype_to_linearize=self.celltype_to_linearize, reverse=self.reverse
        )
        print(cell_index.shape)
        G = self.create_graph(A, gi, self.g, self.threshold_a, self.threshold_b)
        # Draw network
        if self.network_node_file:
            network_node = pd.read_csv(self.network_node_file, sep='\t', index_col=0)
            nodes_to_remove = [n for n in G.nodes if n not in network_node.index]
            G.remove_nodes_from(nodes_to_remove)
            self.fix_pos = {n: [network_node.loc[n, 'x_position'], 1 - network_node.loc[n, 'y_position']] 
                   for n in G.nodes}
        elif self.fix_pos is None:
            self.fix_pos = nx.spring_layout(G)
        
        pos = self.fix_pos

        nx.draw_networkx_nodes(G, pos, ax=self.ax_network, node_color='#EAE0D5', node_size=1500)
        nx.draw_networkx_labels(G, pos, ax=self.ax_network, font_size=10, font_color='#0A0908')

        for u, v, data in G.edges(data=True):
            color = '#A3320B' if data['weight'] > 0 else '#61988E'
            width = abs(data['weight']) * 5
            nx.draw_networkx_edges(G, pos, ax=self.ax_network, edgelist=[(u, v)], width=width, edge_color=color, 
                                   arrows=True, arrowsize=10, connectionstyle=f"arc3,rad={0.10}", node_size=1500)

        self.ax_network.set_title(f"Time: {aptime:.4f}")

        if self.show_scatter:
            # Update scatter plot
            self.adata_subset.obs['changed_clusters'] = pd.Categorical(
                [pd.NA] * len(self.adata_subset),
                categories=self.adata_subset.obs[self.cell_label].cat.categories
            )
            for cell in cell_index:
                if cell in self.adata_subset.obs_names:
                    self.adata_subset.obs.at[cell, 'changed_clusters'] = self.adata_subset.obs.at[cell, self.cell_label]
            
            has_umap, has_tsne = 'X_umap' in self.adata.obsm, 'X_tsne' in self.adata.obsm

            if has_umap:
                scv.pl.umap(self.adata_subset, color='changed_clusters', ax=self.ax_scatter, 
                            title='Network Cells', show=False, legend_loc='right margin')
            if has_tsne:
                scv.pl.tsne(self.adata_subset, ax=self.ax_scatter, 
                            title='Network Cells', show=False, legend_loc='right margin')
            else:
                print("No UMAP or TSNE embedding. Consider computing it first")

        return self.ax_network, self.ax_scatter


import torch
from torch.utils.data import Dataset
import numpy as np
from anndata import AnnData

def split_data(
    adata: AnnData,
    percent: float,
    sample_obs: str,
    layer = ['spliced','unspliced']
):
    """
    Split the dataset for training and validation

    Parameters
    ----------
    adata
        The `AnnData` object for the whole dataset
    percent
        The percentage to be used for training the model
    sample_obs
        The column of sample/treatment index
    layer
        Layer used for training
        (Default: ['spliced','unspliced'])
    Returns
    ----------
    `AnnData` object for training and validation
    """
    n_cells = adata.n_obs
    n_train = int(np.ceil(n_cells * percent))
    n_val = n_cells-n_train
    
    indices = np.random.permutation(n_cells)
    train_idx = np.random.choice(indices, n_train, replace = False)
    indices2 = np.setdiff1d(indices, train_idx)
    val_idx = np.random.choice(indices2, n_val, replace = False)


    train_data = [adata.layers[layer[0]][train_idx], adata.layers[layer[1]][train_idx], adata.obs['latent_time'][train_idx].values, adata.obs[sample_obs][train_idx].values]
    val_data = [adata.layers[layer[0]][val_idx], adata.layers[layer[1]][val_idx], adata.obs['latent_time'][val_idx].values, adata.obs[sample_obs][val_idx].values]
    return train_data, val_data


class MakeDataset(Dataset):
    """
    A class to generate Dataset

    Parameters
    ----------
    adata
        An `AnnData` object
    """

    def __init__(
        self,
        su_data,
    ):

        self.data = su_data
        self.data[0] = torch.Tensor(self.data[0])
        self.data[1] = torch.Tensor(self.data[1])
        self.data[2] = torch.Tensor(self.data[2])
        self.data[3] = torch.Tensor(self.data[3])

    def __len__(self):
        return self.data[0].size(0)

    def __getitem__(self, idx):
        return self.data[0][idx], self.data[1][idx], self.data[2][idx], self.data[3][idx]



from setuptools import setup, find_packages

REQUIRED = [
    'numpy==1.23.5',
    'scikit-learn==1.3.0',
    'pandas==1.5.0',
    'scipy==1.11.3',
    'seaborn==0.12.2',
    'matplotlib==3.7.2',
    'tqdm==4.66.1',
    'scanpy==1.9.4',
    'anndata==0.9.2',
    'scvelo==0.2.5',
    'statsmodels==0.14.0',
    'torch==2.0.0',
    'torchdiffeq==0.2.3',
    'numba==0.56.4',
    'networkx==3.1',
]

setup(
    name='neurovelo',
    version='1.2.0',
    author='Idris Kouadri Boudjelthia',
    author_email='ikouadri@sissa.it',
    description='Interpretable learning of cellular dynamics and gene interactions network',
    packages=find_packages(),
    install_requires=REQUIRED,
    python_requires='>=3.8, <3.10',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
)
